#include <iostream>
#include <fstream>
#include <cmath>
#include <string>
#include <sstream>

#include "lineInter.h"
// ,   -   .
bool TableFunction::calculationLinearApproximation(double *A, double *B)
{
    if (number < 2)
        return false;

    //     y = Ax + B

    double summX = 0;
    double summX_2 = 0;
    double summY = 0;
    double summXY = 0;

    for (int i = 0; i < this->number; i++)
    {
        summX += this->values[i].x;
        summX_2 += summX * summX;
        summY += this->values[i].y;
        summXY += summX * summY;
    }

    //  A  B    :
    // A * summX_2 + B * summX = summXY
    // A * summX + B * number = summY

    // 
    // A = delta1 / delta
    // B = delta2 / delta
    // A = (summXY * number - summX * summY) / (summX_2 * number - summX * summX)
    // B = (summX_2 * summY - summXY * summX) / (summX_2 * number - summX * summX)

    double delta = summX_2 * number - summX * summX;
    *A = (summXY * number - summX * summY) / delta;
    *B = (summX_2 * summY - summXY * summX) / delta;

    return true;
}

TableFunction::TableFunction()
{
    this->a = 0;
    this->b = 0;
    this->number = 1;

    this->values = new Point[number];
    this->values[0].x = 0;
    this->values[0].y = 0;
}

// 
TableFunction::TableFunction(double a, double b, int number)
{
    if (a > b)
    {
        double temp = a;
        a = b;
        b = temp;
    }

    this->a = a;
    this->b = b;
    this->number = number;

    this->values = new Point[number];
    for (int i = 0; i < number; i++)
    {
        this->values[i].x = 0;
        this->values[i].y = 0;
    }
}

//  
TableFunction::TableFunction(const TableFunction &cpy)
{
    this->a = cpy.a;
    this->b = cpy.b;
    this->number = cpy.number;

    this->values = new Point[number];
    for (int i = 0; i < number; i++)
    {
        this->values[i].x = cpy.values[i].x;
        this->values[i].y = cpy.values[i].y;
    }
}

TableFunction & TableFunction::operator= (const TableFunction &cpy)
{
    //    -  
    if (this == &cpy)
        return *this;

    this->a = cpy.a;
    this->b = cpy.b;
    this->number = cpy.number;

    delete []values;
    this->values = new Point[number];
    for (int i = 0; i < number; i++)
    {
        this->values[i].x = cpy.values[i].x;
        this->values[i].y = cpy.values[i].y;
    }
}

TableFunction & TableFunction::operator+= (const TableFunction &g)
{
    if (this->a != g.a || this->b != g.b || this->number != g.number)
    {
        std::cout << "  !\n";
        return *this;
    }

    for (int i = 0; i < this->number; i++)
    {
        this->values[i].y += g.values[i].y;
    }
    return *this;
}

TableFunction & TableFunction::operator-= (const TableFunction &g)
{
    if (this->a != g.a || this->b != g.b || this->number != g.number)
    {
        std::cout << "  !\n";
        return *this;
    }

    for (int i = 0; i < this->number; i++)
    {
        this->values[i].y -= g.values[i].y;
    }
    return *this;
}

TableFunction & TableFunction::operator*= (const TableFunction &g)
{
    if (this->a != g.a || this->b != g.b || this->number != g.number)
    {
        std::cout << "  !\n";
        return *this;
    }

    for (int i = 0; i < this->number; i++)
    {
        this->values[i].y *= g.values[i].y;
    }
    return *this;
}

void TableFunction::operator+= (double indent)
{
    for (int i = 0; i < this->number; i++)
    {
        this->values[i].y += indent;
    }
}

void TableFunction::operator-= (double indent)
{
    for (int i = 0; i < this->number; i++)
    {
        this->values[i].y -= indent;
    }
}

void TableFunction::operator*= (double indent)
{
    for (int i = 0; i < this->number; i++)
    {
        this->values[i].y *= indent;
    }
}

void TableFunction::print()     const
{
    for (int i = 0; i < number; i++)
    {
        std::cout << this->values[i].x << '\t' << this->values[i].y << '\n';
    }
}

double TableFunction::getDX()   const
{
    if (number == 1)
        return 0;

    return (this->b - this->a) / (number - 1);
}

bool TableFunction::loadFromFile(const char *fileName)
{
    std::ifstream fileCin;
    fileCin.open(fileName);
    if (fileCin == 0)
        return false;

    Point newPoints[512];
    int i = 0;
    while (!fileCin.eof())//   ,  .
    {
        double newX;
        fileCin >> newX;

        if (fileCin.eof())
        {
            fileCin.close();
            return false;
        }

        double newY;
        fileCin >> newY;

        newPoints[i].x = newX;
        newPoints[i].y = newY;

        i++;
    }

    if (values)
        delete []values;

    this->number = i;
    this->a = newPoints[0].x;
    this->b = newPoints[number-1].x;
    this->values = new Point[number];

    for (int i = 0; i < number; i++)
    {
        this->values[i].x = newPoints[i].x;
        this->values[i].y = newPoints[i].y;
    }

    fileCin.close();
    return true;
}

bool TableFunction::saveToFile(const char *fileName)
{
    std::ofstream fileCout;
    fileCout.open(fileName);
    for (int i = 0; i < number; i++)
    {
        fileCout << this->values[i].x << '\t' << this->values[i].y << '\n';
    }
    fileCout.close();
}

std::string TableFunction::getFunctionOfLinearApproximation()
{
    double A;
    double B;

    if (calculationLinearApproximation(&A, &B) == false)
        return "error";

    std::stringstream ss;

    if (A == 0 && B == 0)
    {
        return "y = 0";
    }

    //  y
    ss << "y = ";

    //   A
    if (A != 0)
    {
        if (A < 0)
        {
            ss << "-";
        }

        if (fabs(A) != 1)
        {
            ss << A;
        }

        ss << "x ";
    }

    //   B
    if (B != 0)
    {
        if (B < 0)
        {
            B *= -1;
            ss << "- ";
        }
        else
        {
            ss << "+ ";
        }

        ss << B;
    }

    return ss.str();
}

bool TableFunction::linearApproximate()
{
    double A;
    double B;

    if (calculationLinearApproximation(&A, &B) == false)
        return false;

    for (int i = 0; i < number; i++)
    {
        this->values[i].y = A * this->values[i].x + B;
    }

    return true;
}

bool TableFunction::linearApproximateAndChangeDX(int number)
{
    if (number < 2)
    {
        return false;
    }

    //      
    //   
    double A;
    double B;

    if (calculationLinearApproximation(&A, &B) == false)
        return false;

    //     
    this->number = number;
    delete []values;
    values = new Point[number];

    double dx = (this->b - this->a) / (number - 1);
    for (int i = 0; i < number; i++)
    {
        values[i].x = a + i * dx;
        values[i].y = A * values[i].x + B;
    }

    return true;
}

// 
TableFunction::~TableFunction()
{
    delete []values;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

TableFunction operator+ (const TableFunction &LHS, const TableFunction &RHS)
{
    TableFunction temp(LHS);
    temp += RHS;
    return temp;
}

TableFunction operator- (const TableFunction &LHS, const TableFunction &RHS)
{
    TableFunction temp(LHS);
    temp -= RHS;
    return temp;
}

TableFunction operator* (const TableFunction &LHS, const TableFunction &RHS)
{
    TableFunction temp(LHS);
    temp *= RHS;
    return temp;
}

TableFunction operator+ (const TableFunction &LHS, double RHS)
{
    TableFunction temp;
    temp += RHS;
    return temp;
}

TableFunction operator- (const TableFunction &LHS, double RHS)
{
    TableFunction temp;
    temp -= RHS;
    return temp;
}

TableFunction operator* (const TableFunction &LHS, double RHS)
{
    TableFunction temp;
    temp *= RHS;
    return temp;
}

TableFunction operator- (const TableFunction &LHS)
{
    TableFunction temp(LHS);
    temp *= -1;
    return temp;
}

std::ostream & operator<< (std::ostream &os, const TableFunction &RHS)
{
    RHS.print();
    return os;
}
